<manifest lang="json">
{
  "type": "a2a-agent",
  "name": "A2A Agent Example App",
  "version": "1.0.0",
  "description": "A2A agent example",
  "a2aAgents": {
      "test-agent": {
          "url": "https://exmaple.com/a2a-agent",
          "headers": {}
      }
  }
}
</manifest> 