<manifest lang="json">
{
  "type": "mcp-server",
  "name": "DeepWiki MCP Server",
  "version": "1.0.0",
  "description": "MCP Server for DeepWiki",
  "mcpServers": {
    "my-server": {
      "type": "streamable-http",
      "url": "https://mcp.deepwiki.com/mcp"
    }
  }
}
</manifest> 