<config lang="json">
{
    "name": "Simple Python App",
    "type": "web-python",
    "version": "0.1.0",
    "description": "A simple Python app",
    "tags": ["python", "data-science", "modern"],
    "ui": "",
    "cover": "",
    "icon": "🐍",
    "api_version": "0.1.8",
    "requirements": ["numpy", "matplotlib"],
    "dependencies": [],
    "startup_config": {
        "timeout": 60,
        "wait_for_service": "default",
        "stop_after_inactive": 600
    }
}
</config>

<script lang="python">
from hypha_rpc import api

class HyphaApp():
    async def setup(self):
        await api.log('initialized')
    
    def add2(self, a):
        return a + 2

api.export(HyphaApp())
</script> 